# This class manages a set of "calibration images" or training data
# each image in the set has a name
#
# operation on this class are:
#   1) create/replace the image for a name
#   2) get the name of the image that best matches an input image
#   3) get a statistical description of the difference between two images
#   4) get a statistical description of a region in an image 
#
# 2) is good for the simple dodgeball:  Am I on the left or right?  
# It is also good for games like simon.
#
# 3+4) is good for continuous dodgeball: Am I at (X,Y)
from PIL import Image, ImageChops, ImageStat;
import math;

class ImageCalibration:
  def __init__(self, listener=0):
    #listener responds to targetTracked(box)
    #and differencesDetected(map)
    self.listener = listener;

    #placeholder
    temp = 0;
    self.imageMap = {};
    self.stddevMap = {}
    self.segmentedImage = 0;
    self.annotatedImage = 0;
    self.centerOfMass = (0, 0);
    #normalized standard deviations away to be interesting
    self.threshold = 1.0; 

  def __getattribute__(self, segmentedImage):
    return self.segmentedImage;

  def __getattribute__(self, centerOfMass):
    return self.centerOfMass;

  def __getattribute__(self, annotatedImage):
    return self.annotatedImage;

  def __getattribute__(self, threshold):
    return self.threshold;

  def getSize(self):
    if (len(self.imageMap) > 0):
      return self.imageMap.values()[0].size;
    else:
      return (0,0);

  def setThreshold(self, threshold):
    #print "setting threshold to: " + str(threshold);
    self.threshold = float(threshold);

  def setImage(self, name, image, keepStats):
    self.imageMap[name] = image; 
    if (keepStats):
      self.stddevMap[name] = ImageStat.Stat(image).stddev;
    else:
      self.stddevMap[name] = (1,1,1);

  def removeImage(self, name):
    self.imageMap.pop(name);
    self.stddevMap.pop(name);

  def getBestMatch(self, image):
    bestName = '';
    bestScore = 1e30;
    for name in self.imageMap.keys():
       sourceImage = self.imageMap[name];
       diff = ImageChops.difference(sourceImage, image); 
       stat = ImageStat.Stat(diff);
       score = 0;
       for dim in [0, 1, 2]:
         if (self.stddevMap[name][dim] != 0):
           score += stat.sum[dim]/self.stddevMap[name][dim];
       if (score < bestScore):
         bestScore = score;
	 bestName = name;
    #print("IC: bestName = " + name + " bestScore = " + str(score));
    return bestName;

  def calc_stats(self, array):
    #returns a tuple of mean, variance
    sum = 0.0;
    #will hold (x-mean)^2 terms
    varAccumulator = 0.0; 
    count = len(array);
    if (count > 0):
      for idx in range(0,count):
        sum += array[idx];
      mean = sum/count;
      for idx in range(0,count):
        varAccumulator += pow((array[idx] - mean), 2);
      std = math.sqrt(varAccumulator/(count - 1));
    else:
      mean = 0;
      std = 0;
    return (mean, std);

  def calculateDifferenceDistribution(self, baseName, targetName, thresh=0.5):
    #does this need error checking
    # do the images exist, etc.
    # what is the right threshold
    baseImage = self.imageMap[baseName];
    targetImage = self.imageMap[targetName];
    diff = ImageChops.difference(baseImage, targetImage); 
    boundingBox = diff.getbbox();
    dRed = [];
    dGreen= [];
    dBlue = [];
    for x in range(boundingBox[0], boundingBox[2]):
      for y in range(boundingBox[1], boundingBox[3]):
        thisDiff = diff.getpixel((x,y));
        totalNormalizedDiff = 0.0;
	for channel in [0, 1, 2]:
	  if (self.stddevMap[baseName][channel] != 0):
	    totalNormalizedDiff += pow((1.0 * thisDiff[channel])/self.stddevMap[baseName][channel], 2);
        if (math.sqrt(totalNormalizedDiff) > thresh):
	  pixel = targetImage.getpixel((x,y));
	  dRed.append(pixel[0]);
	  dGreen.append(pixel[1]);
	  dBlue.append(pixel[2]);
    rStats = self.calc_stats(dRed);
    gStats = self.calc_stats(dGreen);
    bStats = self.calc_stats(dBlue);
    self.diffMeans = [rStats[0], gStats[0], bStats[0]];
    self.diffStds = [rStats[1], gStats[1], bStats[1]];
    print("Means: " + str(self.diffMeans));
    print("Standard Deviations: " + str(self.diffStds));

  def old_calculateImageMap(self, baseName, targetImage, boundingBox=0):
    if (boundingBox == 0):
      boundingBox = (0,0)+targetImage.size;
    baseImage = self.imageMap[baseName];
    mapImage = Image.new("RGB", baseImage.size);
    diffImage = ImageChops.difference(baseImage, targetImage);
    baseMarker = (255,255,255);
    trackMarker = (0, 100, 225);
    for x in range(boundingBox[0], boundingBox[2]):
      for y in range(boundingBox[1], boundingBox[3]):
        baseDiff = 0.0;
	targetDiff = 0.0;
	diffPixel = diffImage.getpixel((x,y));
	targetPixel = targetImage.getpixel((x,y));
	for channel in [0, 1, 2]:
	  baseDiff += pow((1.0 * diffPixel[channel])/self.stddevMap[baseName][channel], 2);
	  targetDiff += pow(1.0 * abs((targetPixel[channel] - self.diffMeans[channel]))/self.diffStds[channel], 2);
	if (math.sqrt(baseDiff) < math.sqrt(targetDiff)):
	  mapImage.putpixel((x,y), baseMarker);
	else:
	  mapImage.putpixel((x,y), trackMarker);
    self.segmentedImage = mapImage;
    return mapImage;

  def old_findCenterOfMass(self, binaryImage, massSize):
    #binary image is the result of calculateImageMap
    #this will search for an area of size massSize(x,y)
    #with large total pixel intensity
    #this probably isn't the most accurate method to use
    #but I think I can implement it quickly, and it will run
    #somewhat well
    #print "Mass size: " + str(massSize);
    mask = Image.new("L", binaryImage.size, 0);
    currentSize = binaryImage.size;
    bestBox = binaryImage.getbbox();
    while ((currentSize[0] > massSize[0]) or (currentSize[1] > massSize[1])):
      #print "---";
      #print "Current: " + str(currentSize);
      #print "Mass: " + str(massSize);
      bestSum = (bestBox[2]-bestBox[0]) * (bestBox[3] - bestBox[1]) * 255;
      currentSize = ((bestBox[2] - bestBox[0]), (bestBox[3] - bestBox[1]));
      currentSize = (int(currentSize[0]/2), int(currentSize[1]/2));
      if (currentSize[0] < massSize[0]):
        currentSize = (massSize[0], currentSize[1]);
      if (currentSize[1] < massSize[1]):
        currentSize = (currentSize[0], massSize[1]);
      boxes = [
      (bestBox[0], bestBox[1], bestBox[0] + currentSize[0], bestBox[1] + currentSize[1]),
      (bestBox[2] - currentSize[0], bestBox[1], bestBox[2], bestBox[1] + currentSize[1]),
      (bestBox[0], bestBox[3] - currentSize[1], bestBox[0] + currentSize[0], bestBox[3]),
      (bestBox[2] - currentSize[0], bestBox[3] - currentSize[1], bestBox[2], bestBox[3]),
      ];#"
      for idx in range(0,4):
       mask.paste(255, boxes[idx]);
       stat = ImageStat.Stat(binaryImage, mask);
       #print boxes[idx];
       #print stat.sum[0];
       # only look at the red channel that's all that's there
       if (stat.sum[0] < bestSum):
         bestBox = boxes[idx];
         bestSum = stat.sum[0];
       mask.paste(0, boxes[idx]);
    #print "best: "
    #print bestBox;
    #print bestSum;
    self.centerOfMass = bestBox;
    if (self.listener):
      self.listener.targetTracked(bestBox);
    return bestBox;


  def calculateImageMap(self, baseName, targetImage):
    baseImage = self.imageMap[baseName];
    self.segmentedImage = Image.new('L',baseImage.size);
    annotatedImage = Image.new('RGB',baseImage.size);
    diff = ImageChops.difference(baseImage, targetImage); 
    mapList = list(diff.getdata());
    annotatedList = list(mapList);
    totalX = 0.0;
    totalY = 0.0;
    totalScore = 0.0;
    maxx = baseImage.size[0];
    idx = 0;
    for y in range(0, baseImage.size[1]):
      for x in range(0, maxx):
        score = 0.0;
        for channel in range(0,3):
	  #score += (l[x+(y*maxx)][channel] * l[x+(y*maxx)][channel]); 
	  temp = mapList[idx][channel] / self.stddevMap[baseName][channel];
	  score = score + (temp * temp);
        score = math.sqrt(score);
	if (score < self.threshold):
	  score = 0;
        score = score * 60;
	totalX += x * score;
	totalY += y * score;
	totalScore += score;
        mapList[idx] = int(score);
        annotatedList[idx] = (0, int(score/2), int(score));
	idx = idx+1;
    if (totalScore > 0):
      comX = int(totalX / totalScore);
      comY = int(totalY / totalScore);
    else:
      #print "no different enough data.";
      comX = -1;
      comY = -1;
    #print "Center of Mass: " + str(comX) + ", " + str(comY);
    self.centerOfMass=(comX, comY);
    self.segmentedImage.putdata(mapList);
    annotatedImage.putdata(annotatedList); 
    #dinner time this was uncommented
    self.annotatedImage = ImageChops.blend(annotatedImage, targetImage, 0.7);
    return self.segmentedImage;

    
